/* KallistiOS ##version##

   flashrom.c
   Copyright (c)2003 Dan Potter
*/

/*

  This module implements the stuff enumerated in flashrom.h. 

  Writing to the flash is disabled by default. To re-enable it, uncomment
  the #define below.

  Thanks to Marcus Comstedt for the info about the flashrom and syscalls.

 */

#include "flashrom.h"
#include <arch/irq.h>

CVSID("$Id: flashrom.c,v 1.4 2003/03/10 01:45:31 bardtx Exp $");

/* This is the fateful define. Re-enable this at the peril of your
   Dreamcast's soul ;) */
/* #define ENABLE_WRITES 1 */


/* First, implementation of the syscall wrappers. */

int flashrom_info(int part, int * start_out, int * size_out) {
	int	(*sc)(int, uint32*, int, int);
	uint32	ptrs[2];
	int	rv;

	*((uint32 *)&sc) = *((uint32 *)0x8c0000b8);
	if (sc(part, ptrs, 0, 0) == 0) {
		*start_out = ptrs[0];
		*size_out = ptrs[1];
		rv = 0;
	} else
		rv = -1;
	return rv;
}

int flashrom_read(int offset, void * buffer_out, int bytes) {
	int	(*sc)(int, void*, int, int);
	int	rv;

	*((uint32 *)&sc) = *((uint32 *)0x8c0000b8);
	rv = sc(offset, buffer_out, bytes, 1);
	return rv;
}

int flashrom_get_region() {
	int start, size;
	char region[6] = { 0 };

	/* Find the partition */
	if (flashrom_info(FLASHROM_PT_SYSTEM, &start, &size) < 0) {
//		dbglog(DBG_ERROR, "flashrom_get_region: can't find partition 0\n");
		return -1;
	}

	/* Read the first 5 characters of that partition */
	if (flashrom_read(start, region, 5) < 0) {
//		dbglog(DBG_ERROR, "flashrom_get_region: can't read partition 0\n");
		return -1;
	}

	/* Now compare against known codes */
	if (!strcmp(region, "00000"))
		return FLASHROM_REGION_JAPAN;
	else if (!strcmp(region, "00110"))
		return FLASHROM_REGION_US;
	else if (!strcmp(region, "00211"))
		return FLASHROM_REGION_EUROPE;
	else {
//		dbglog(DBG_WARNING, "flashrom_get_region: unknown code '%s'\n", region);
		return FLASHROM_REGION_UNKNOWN;
	}
}


